#include "StdAfx.h"
#include "haptics.h"
#include "scene.h"



#ifdef HAPTICS_CHAI
	cMeta3dofPointer *tool=NULL;
	cWorld *world=NULL;
	cPrecisionTimer hapticTimer;
	
	
	
	void callbackContact(void *unusedParam) {
		tool->updatePose();
		#ifdef USE_ROBOTIC_WORLD
			callbackContact(reinterpret_cast<const double*>(&tool->m_deviceGlobalPos),reinterpret_cast<double*>(&tool->m_lastComputedGlobalForce));
		#else
			double position[3]={tool->m_deviceGlobalPos[1],tool->m_deviceGlobalPos[2],tool->m_deviceGlobalPos[0]};
			double force[3];
			callbackContact(position,force);
			tool->m_lastComputedGlobalForce.set(force[2],force[0],force[1]);
		#endif
		tool->applyForces();
	}
#endif
#ifdef HAPTICS_HDAL
	HDLDeviceHandle deviceHandle=HDL_INVALID_HANDLE;
	HDLOpHandle callbackHandle=HDL_INVALID_HANDLE;
	double deviceWorkspace[6];
	double workspace[6]={-WORKSPACE_SIZE,-WORKSPACE_SIZE,-WORKSPACE_SIZE,WORKSPACE_SIZE,WORKSPACE_SIZE,WORKSPACE_SIZE};
	double transformFromDevice[16];
	double toolPositionServo[3];
	double toolForceServo[3],toolForce[3];



	HDLServoOpExitCode callbackContact(void *unusedParam) {
		hdlToolPosition(toolPositionServo);
		double position[3];
		deviceToWorld(toolPositionServo,position);
		callbackContact(position,toolForceServo);
		hdlSetToolForce(toolForceServo);
		return HDL_SERVOOP_CONTINUE;
	}
	//--------------------------------------------------------------------------------------------------
	HDLServoOpExitCode callbackSynchronize(void *unusedParam) {
		hdlToolPosition(toolPositionServo);
		deviceToWorld(toolPositionServo,toolPosition);
		thickness=thicknessServo;
		return HDL_SERVOOP_EXIT;
	}
	//--------------------------------------------------------------------------------------------------
	void deviceToWorld(const double device[3],double world[3]) {
		world[0]=(
			transformFromDevice[0]*device[0]
			+transformFromDevice[4]*device[1]
			+transformFromDevice[8]*device[2]
			+transformFromDevice[12]
		);
		world[1]=(
			transformFromDevice[1]*device[0]
			+transformFromDevice[5]*device[1]
			+transformFromDevice[9]*device[2]
			+transformFromDevice[13]
		);
		world[2]=(
			transformFromDevice[2]*device[0]
			+transformFromDevice[6]*device[1]
			+transformFromDevice[10]*device[2]
			+transformFromDevice[14]
		);
	}
#endif



GLfloat thicknessServo=0.0f,thickness=0.0f;
double toolPosition[3];



void initHaptics() {
	#ifdef HAPTICS_CHAI
		world=new cWorld();
		tool=new cMeta3dofPointer(world);
		tool->initialize();
		tool->setWorkspace(2*WORKSPACE_SIZE,2*WORKSPACE_SIZE,2*WORKSPACE_SIZE);
		tool->computeGlobalPositions();
		tool->start();
		tool->setForcesON();
		hapticTimer.set(1,&callbackContact,NULL);
	#endif
	#ifdef HAPTICS_HDAL
		deviceHandle=hdlInitNamedDevice(NULL);
		hdlStart();
		hdlMakeCurrent(deviceHandle);
		hdlDeviceWorkspace(deviceWorkspace);
		hdluGenerateHapticToAppWorkspaceTransform(deviceWorkspace,workspace,true,transformFromDevice);
		callbackHandle=hdlCreateServoOp(&callbackContact,NULL,false);
	#endif
}
//--------------------------------------------------------------------------------------------------
void uninitHaptics() {
	#ifdef HAPTICS_CHAI
		hapticTimer.stop();
		if (tool!=NULL) {
			tool->stop();
			delete tool;
		}
		if (world!=NULL) delete world;
	#endif
	#ifdef HAPTICS_HDAL
		if (callbackHandle!=HDL_INVALID_HANDLE) {
			hdlDestroyServoOp(callbackHandle);
			callbackHandle=HDL_INVALID_HANDLE;
		}
		hdlStop();
		if (deviceHandle!=HDL_INVALID_HANDLE) {
			hdlUninitDevice(deviceHandle);
			deviceHandle=HDL_INVALID_HANDLE;
		}
	#endif
}
//--------------------------------------------------------------------------------------------------
void updateTool() {
	#ifdef HAPTICS_CHAI
		thickness=thicknessServo;
		#ifndef USE_ROBOTIC_WORLD
			toolPosition[0]=tool->m_deviceGlobalPos[1];
			toolPosition[1]=tool->m_deviceGlobalPos[2];
			toolPosition[2]=tool->m_deviceGlobalPos[0];
		#endif
	#endif
	#ifdef HAPTICS_HDAL
		hdlCreateServoOp(&callbackSynchronize,NULL,true);
	#endif
}
//--------------------------------------------------------------------------------------------------
void callbackContact(const double position[3],double force[3]) {
	double dist2=position[0]*position[0]+position[1]*position[1]+position[2]*position[2];
	if (dist2>SPHERE_RADIUS*SPHERE_RADIUS) {
		force[0]=force[1]=force[2]=0.0;
		thicknessServo=0.0f;
		return;
	}
	double dist=sqrt(dist2);
	if (dist>SMALL_SPHERE_RADIUS) {
		double ramp;
		if (dist>=0.9*SPHERE_RADIUS) {
			ramp=(SPHERE_RADIUS-dist)*(1.0/(0.1*SPHERE_RADIUS));
		} else {
			ramp=1.0;
		}
		double distInv=1.0/dist;
		thicknessServo=1.0f-GLfloat((dist-SMALL_SPHERE_RADIUS)*(1.0/(SPHERE_RADIUS-SMALL_SPHERE_RADIUS)));
		force[0]=ramp*STIFFNESS1*position[0]*distInv;
		force[1]=ramp*STIFFNESS1*position[1]*distInv;
		force[2]=ramp*STIFFNESS1*position[2]*distInv;
		return;
	} else {
		thicknessServo=1.0f;
		if (abs(dist)<1e-6) {
			force[0]=force[2]=0.0;
			force[1]=10.0;
			return;
		}
		double distInv=1.0/dist;
		double pene=(SMALL_SPHERE_RADIUS-dist)*(1.0/SMALL_SPHERE_RADIUS);
		force[0]=(STIFFNESS1+pene*STIFFNESS2)*position[0]*distInv;
		force[1]=(STIFFNESS1+pene*STIFFNESS2)*position[1]*distInv;
		force[2]=(STIFFNESS1+pene*STIFFNESS2)*position[2]*distInv;
		return;
	}
}
//--------------------------------------------------------------------------------------------------
const double* getToolPosition() {
	#ifdef HAPTICS_CHAI
		#ifdef USE_ROBOTIC_WORLD
			return reinterpret_cast<const double*>(&tool->m_deviceGlobalPos);
		#else
			return toolPosition;
		#endif
	#endif
	#ifdef HAPTICS_HDAL
		return toolPosition;
	#endif
}
